!===============================================================
!
!  PROGRAM: Cell_Centered
!
!  PURPOSE: Unsteady Heat Conduction
!
!  DATE:	1/10/04
!===============================================================
!

Program Cell_Centered
USE FE_mesh_IO
Implicit none

! Inputs
Integer :: n_power, k_num, mode
Character (LEN=25) :: filename
Real*8	:: dt, final_time, alpha
! Work 
Real*8	:: time, pi = 3.141592653589, sav, t1,t2

! Geometry
Real*8, Allocatable, Dimension(:,:) :: fpos, cpos, norm, rx1,rx2
Real*8, Allocatable, Dimension(:) :: volci, AoL

! variables
Real*8, Allocatable, Dimension(:) :: T


! set up
    Call Get_inputs()
!!    Call Set_inputs()

! read in the mesh
    Call Read_FV(trim(filename))

! compute geometry 
    Allocate (  fpos(ND,NF), cpos(ND,0:NC), norm(ND,NF), volci(0:NC), AoL(NF), rx1(ND,NF),rx2(ND,NF) )
    Call Geom_FV(mode)  

! Guts start here
! initialize solution
    Allocate ( T(0:NC) )
	Call Init_FV()

! solve
    time = 0.0
	if (n_power > 0) Call Write_Moment()

    Call CPU_TIME(t1)
    do while (time < final_time - dt*0.99)
		time = time + dt

		if (mode == 0) then 
          Call Update_FV_CC()
		else 
		  Call Update_FV_CG()
		end if

		if (n_power > 0) Call Write_Moment()
	end do
	Call CPU_TIME(t2)
	write(*,*) 'Number of cells = ',NC
    write(*,*) 'CPU_TIME, CPU_time per cell = ',t2-t1, (1.e6)*(t2-t1)*dt/time/Real(NC)
    Call Write_Err()    

! Clean up 
	Call Write_Tecplot(trim(filename))  
	Call Deallocate_Init_FV()			! variables
	Call Deallocate_Geom_FV()			! geometry

	Call Deallocate_FV_Con ()           ! mesh
	Call Deallocate_FV_Pos ()
	Call Deallocate_FV_Mat ()
 
		    
Stop


contains

!===============================================================
Subroutine Get_inputs
Implicit None

    write(*,*) 'What is the Mesh File name ? (without .FV extension)'
	read(*,*) filename

    write(*,*) 'mode, dt, final_time, alpha, k, n'
	read(*,*) mode, dt, final_time, alpha, k_num, n_power

Return
End Subroutine Get_inputs

!===============================================================
Subroutine Set_inputs
Implicit None

	filename = "mesh/square_reg_2k"

	mode = 1     !0 for CC, 1 for CG    
	final_time = 1.0
	alpha = 0.01
    n_power = 0
	k_num = 1
	dt = 0.01/alpha/REAL(25000*(k_num**2))


Return
End Subroutine Set_inputs

!===============================================================

Subroutine Read_FV(filename)
Implicit None
Character (Len=*) :: filename

Integer :: unit,ios

    unit = 10
    Open (unit, File=filename//".FV", Iostat=ios, Form="Formatted", Action="Read")
	if (ios /= 0) then
		Print *, "Error opening .FV file. Error code = ", ios, filename//".FV" 
		stop 
	end if

	Read(unit,*) 
	Read(unit,*) ND
	Read(unit,*) 

	Call Read_FV_Con (unit)
	Call Read_FV_Pos (unit) 
	Call Read_FV_Mat (unit)

	Close (unit)
	Write(*,*) "Read done"

Return
End Subroutine Read_FV


!===============================================================

Subroutine Geom_FV(id)
Implicit None
Integer :: id

Integer :: i,n, n1,n2,c1,c2
Real*8	:: mov(ND,0:NC)

 !   Allocate (  fpos(ND,NF), cpos(ND,0:NC), norm(ND,NF), volci(0:NC), AoL(NF), rx1(ND,NF),rx2(ND,NF) )


	cpos = 0.0
    do i=1,NF
	  n1 = F2N(1,i)
	  n2 = F2N(2,i)
	  c1 = F2C(1,i)
	  c2 = F2C(2,i)
      fpos(:,i) = 0.5*( npos(:,n2) + npos(:,n1) )
	  norm(2,i) = npos(1,n2) - npos(1,n1)
	  norm(1,i) = npos(2,n1) - npos(2,n2)
	  cpos(:,c1) = cpos(:,c1) + fpos(:,i)
	  cpos(:,c2) = cpos(:,c2) + fpos(:,i)
	end do
	cpos = cpos/3.0
	cpos(:,0) = 0.0

! volume using CG
    volci = 0.0
    do i=1,NF
	  c1 = F2C(1,i)
	  c2 = F2C(2,i)

      rx1(:,i) = fpos(:,i)- cpos(:,c1) 
      rx2(:,i) = fpos(:,i)- cpos(:,c2) 	  
	  if (c2 == 0) rx2(:,i) = 0.0

      volci(c1) = volci(c1) + 0.5*ABS(dot_product( rx1(:,i),norm(:,i) ))
      volci(c2) = volci(c2) + 0.5*ABS(dot_product( rx2(:,i),norm(:,i) ))

	  AoL(i) = dot_product( rx1(:,i)-rx2(:,i) , norm(:,i) )
	  norm(:,i) = norm(:,i)*sign( 1.0d0, AoL(i) )
	end do
	WHERE(volci /= 0.0) volci = 1.0/volci
	volci(0) = 1.0

if (id == 0) then  ! change cpos to circumcenter
do n=1,40

    mov = 0.0
    do i=1,NF
	  c1 = F2C(1,i)
	  rx1(:,i) = fpos(:,i) - cpos(:,c1)
      mov(:,c1) = mov(:,c1) + rx1(:,i) - norm(:,i)*dot_product(rx1(:,i),norm(:,i))/dot_product(norm(:,i),norm(:,i)) 

	  c2 = F2C(2,i)
	  rx2(:,i) = fpos(:,i) - cpos(:,c2)
      mov(:,c2) = mov(:,c2) + rx2(:,i) - norm(:,i)*dot_product(rx2(:,i),norm(:,i))/dot_product(norm(:,i),norm(:,i)) 
    end do
	mov(:,0) = 0.0
	cpos = cpos + 0.5*mov 
!	write(*,*)'CC ',SQRT( dot_product(mov(1,:),mov(1,:))+dot_product(mov(2,:),mov(2,:))/REAL(NC)**2 )

end do

! other stuff
    do i=1,NF
	  c1 = F2C(1,i)
	  c2 = F2C(2,i)

      rx1(:,i) = fpos(:,i)- cpos(:,c1) 
      rx2(:,i) = fpos(:,i)- cpos(:,c2) 	  
	  if (c2 == 0) rx2(:,i) = 0.0

      AoL(i) = dot_product( norm(:,i),norm(:,i) ) / ( dot_product( rx1(:,i)-rx2(:,i) , norm(:,i) ) )
	  if (AoL(i) < 0.0) write(*,*) 'NEG face ',i 
	end do

end if


Return
End Subroutine Geom_FV

!
!===============================================================
!
Subroutine Deallocate_Geom_FV ()
Implicit None

    Deallocate (  fpos, cpos, norm, volci, AoL, rx1,rx2 )    

Return
End Subroutine Deallocate_Geom_FV

!===============================================================

Subroutine Init_FV()
Implicit None

Integer :: i

    forall (i=1:NC) T(i) = sin(pi*REAL(k_num)*cpos(1,i))*sin(pi*REAL(k_num)*cpos(2,i))
	T(0) = 0.0

Return
End Subroutine Init_FV


!
!===============================================================
!
Subroutine Deallocate_Init_FV ()
Implicit None

    Deallocate (  T )	  

Return
End Subroutine Deallocate_Init_FV
!
!===============================================================
!
Subroutine Update_FV_CC()
Implicit None

Integer :: i, c1,c2
Real*8:: Diff(NF), tmp(0:NC), flux


    Diff  = (dt*alpha)*AoL
	tmp = 0.0
	T(0) = 0
    do i=1,NF
	  c1 = F2C(1,i)
	  c2 = F2C(2,i)
      flux = (T(c2) - T(c1))*Diff(i)
	  tmp(c1) = tmp(c1) + flux
      tmp(c2) = tmp(c2) - flux
	end do

    T = T + tmp*volci 
	T(0) = 0.0 

Return
End Subroutine Update_FV_CC

!===============================================================
!
Subroutine Update_FV_CG()
Implicit None

Integer :: i, c1,c2
Real*8:: gtmp(ND,0:NC), tmp(0:NC), flux, dx(2,NF), sum


	gtmp = 0.0
	tmp = 0.0
	T(0) = 0
    do i=1,NF
	  c1 = F2C(1,i)
	  c2 = F2C(2,i)
      dx(1,i) = 1.0/volci(c1) !dot_product(rx1(:,i), norm(:,i))
      dx(2,i) = 1.0/volci(c2) !dot_product(rx2(:,i), norm(:,i))
	  if (c2 == 0) dx(2,i) = 0.0

      sum = dx(1,i) + dx(2,i)
	  dx(1,i) = dx(1,i)/sum
	  dx(2,i) = dx(2,i)/sum

      flux = (T(c2) - T(c1))
	  gtmp(:,c1) = gtmp(:,c1) + flux*norm(:,i)*dx(1,i)
      gtmp(:,c2) = gtmp(:,c2) + flux*norm(:,i)*dx(2,i)
	end do
    
    forall(i=1:ND) gtmp(i,:) = gtmp(i,:)*(dt*alpha)*volci(:)

    do i=1,NF
	  c1 = F2C(1,i)
	  c2 = F2C(2,i)
      flux = dot_product( norm(:,i), gtmp(:,c1)*dx(1,i)+gtmp(:,c2)*dx(2,i) ) 
	  tmp(c1) = tmp(c1) + flux
      tmp(c2) = tmp(c2) - flux
	end do

    T = T + tmp*volci 
	T(0) = 0.0 

Return
End Subroutine Update_FV_CG

!===============================================================

Subroutine Write_Moment()
Implicit None
Real*8	:: ext, tmp

    T(0) = 0.0
    ext = alpha*pi*pi*2.0*k_num*k_num*REAL(n_power*n_power+n_power)
    if (time == 0.0) then
	  sav = SUM( (T**(n_power+1))/volci )
      Write(14,*)time,  -ext
 
    else
      tmp = SUM( (T**(n_power+1))/volci ) 
      Write(14,*)time, (tmp - sav) / (tmp + sav) / (0.5*dt)
      sav = tmp
    end if

Return
End Subroutine Write_Moment

!===============================================================

Subroutine Write_Err()
Implicit None
Real*8	:: lam, Terr(0:NC)
Integer :: i

	lam = pi*pi*alpha*2.0*Real(k_num*k_num)

    forall (i=1:NC) Terr(i) = T(i) - sin(pi*REAL(k_num)*cpos(1,i))*sin(pi*REAL(k_num)*cpos(2,i))*exp(-lam*time)
	Terr(0) = 0.0 
	volci(0) = 1.0
		 
	lam = MAXVAL(Terr)  
    Write(*,*)SUM(ABS(Terr))/Real(NC), SQRT( SUM(Terr*Terr)/Real(NC) ), MAXVAL(ABS(Terr)) ! , MAXLOC(ABS(Terr))
 !   Write(*,*)SUM(ABS(Terr)/volci), SQRT( SUM(Terr*Terr/volci) ), MAXVAL(ABS(Terr))

Return
End Subroutine Write_Err


!===============================================================

Subroutine Write_Tecplot(filename)
Implicit None
Character (Len=*) :: filename

Integer :: unit,ios
Integer :: i, c1,c2
Character (LEN=2) :: inc

		unit = 10
		Open (unit, File=filename//".dat", Iostat=ios, Form="Formatted", Action="Write")
		if (ios /= 0) then
			Print *, "Error opening .dat file. Error code = ", ios, filename//".dat" 
			stop 
		end if

		do i=1,NC
			write(10,*) cpos(1,i),cpos(2,i),T(i)
		end do

        close(10)

Return
End Subroutine Write_Tecplot



!===============================================================
!===============================================================
!===============================================================
End Program Cell_Centered

